#!/opt/local/bin/perl

use strict;
use Math::Combinatorics;

our $pi = atan2(1,1) *4;
our $degToRad = $pi / 180.0;
our $evToKjmol = 96.485; # eV to kJ/mol

# Temperature
our $T = $ARGV[0];
print "Temperature = $T K\n\n";

# Distance of route a and tilde c
my $dtc_a = 2.746e-10; # d_a in paper
my $dtc_c = 2.661e-10; # d_{\tilde c} in paper

# Angle between route a and route c
my $angle_ac = 58.937590529475 * $degToRad; # \theta_{a \tilde c} in paper
my $angle_cc = 31.0624094705 * $degToRad;   # \theta_{c \tilde c} in paper

# Vacancy formation energy at top and base sites
my $ve_top = 0.268816;
my $ve_base= 0.225016;

# Barrier energy of ion jump through diffusion routes
# be_f is for route 1-5 and be_r is for route 1', 3'
my @be_f; my @be_r;

# Vib. freq. of hopping ion in jump direction.
my @fq_f; my @fq_r;

# Jump frequency
my @eta_f; my @eta_r; # \eta in paper

# Distances of diffusion routes
my @dtc; # d in paper

# Input barrier energies and vib. frequencies [sec^{-1}]
$be_f[1]  = 0.56574766; $fq_f[1]  = 3675321702068.44;  
$be_r[1]  = 0.60965692; $fq_r[1]  = 3574830738373.0 ; 
$be_f[2]  = 0.72189905; $fq_f[2]  = 4151680469793.08; 
$be_f[3]  = 0.22475453; $fq_f[3]  = 3154907260113.56; 
$be_r[3]  = 0.26855453; $fq_r[3]  = 2541783820157.8 ;
$be_f[4]  = 0.6238511 ; $fq_f[4]  = 4137620857992.41; 
$be_f[5]  = 0.53482665; $fq_f[5]  = 3617777758588.7 ; 

# Input distance of diffusion routes [meter]
$dtc[1]  = 2.6605e-10;
$dtc[2]  = 2.6360e-10;
$dtc[3]  = 2.6600e-10;
$dtc[4]  = 2.7943e-10;
$dtc[5]  = 2.7457e-10;

# Calculate jump frequency
for my $i (1,2,3,4,5){
    $eta_f[$i] = $fq_f[$i]*&BF($be_f[$i]);
    $eta_r[$i] = $fq_r[$i]*&BF($be_r[$i]);
}

# Display quantities related to ion jump
print "Route | Barrier E [eV, kJ/mol] | Vib. freq. [sec^{-1}] | Jump freq. [sec^{-1}] \n";
for my $i (1,2,3,4,5) {
    printf "%3d | %7.3f , %7.3f | %.3e | %.3e \n", $i, $be_f[$i], $be_f[$i]*$evToKjmol, $fq_f[$i], $eta_f[$i];
    if ($i==1 || $i==3){
	printf "%3d | %7.3f , %7.3f | %.3e | %.3e \n", $i, $be_r[$i], $be_r[$i]*$evToKjmol, $fq_r[$i], $eta_r[$i];
    }	
}
print "\n";

# Vacancy formation rates at top and base sites and rep-site.
my $Cv_top = &BF($ve_top);     # C_v^{(top)}  in paper
my $Cv_base= &BF($ve_base);    # C_v^{(base)} in paper
my $Cv = $Cv_top + 2*$Cv_base; # C_v^{(rep)}  in paper

my $gamma_top = $Cv_top  / $Cv;   # \gamma^{(top)}  in paper
my $gamma_beta  = $Cv_base / $Cv; # \gamma^{(base)} in paper

print "     Vacancy formation rate | formation energy [eV, kJ/mol]\n";
printf "    Top : %.3e | %.3f, %.2f \n", $Cv_top, $ve_top, $ve_top *$evToKjmol;
printf "    Base: %.3e | %.3f, %.2f \n", $Cv_base,$ve_base,$ve_base*$evToKjmol;
printf "    Rep : %.3e | N/A\n\n", $Cv;

print "Probability that a vacancy found in top/base site\n";
print "after a vacancy has moved in a rep-site numerous times\n";
printf "    gamma_top(top)  = %5.2f \%\n",    $gamma_top  * 100;
printf "    gamma_beta(base) = %5.2f \%\n\n", $gamma_beta * 100;

# jump frequency of jump s/t;
my $eta_a =  $gamma_beta*$eta_f[5]; #\eta_a in paper
my $eta_c =  $gamma_top*$eta_f[1]  + $gamma_beta*($eta_r[1]+$eta_f[2]);
                                    #\eta_{\tilde c} in paper
my $rate_a = $eta_a/(2*$eta_a+4*$eta_c); # p_a in paper
my $rate_c = $eta_c/(2*$eta_a+4*$eta_c); # p_{\tilde c} in paper

print  "rate_a: $rate_a\n";
print  "rate_c: $rate_c\n";
printf "rates/ratet = %.4f\n\n", $rate_a/$rate_c;

my $sum_even =  &back_even($rate_a, 50);        # \pi_{even}    in paper 
my $sum_odd  =  &back_odd($rate_a, 51);         # \pi_{odd}     in paper
my $sum_oneside  =  &back_oneside($rate_a, 50);  # \pi_{oneside} in paper

my @cos_a; my @cos_c; # <\cos \theta>_{a, \tilde c}^{(n'=1,2,3,...)}

# Solve simultaneous recurrence formula of eq.6,7 in supplementary information
$cos_a[0] = 1; $cos_c[0] = 1;
for my $i (1..100) {
    $cos_a[$i] = $sum_oneside * $rate_a * (-1) * $cos_a[$i-1]
	+ 2 * $sum_oneside * $sum_even *  $rate_c * $rate_c * ($dtc_c/$dtc_a)**1 * (-cos($angle_ac)) * $cos_c[$i-1];
    $cos_c[$i] = $sum_even * $rate_c * (-1) * $cos_c[$i-1] 
	+ ($sum_even+$sum_odd) * $sum_oneside * $rate_c * $rate_a * ($dtc_a/$dtc_c)**1 * (-cos($angle_ac)) * $cos_a[$i-1]
	+ $sum_odd * $rate_c * (-cos(2.*$angle_cc)) * $cos_c[$i-1];
	+ $sum_odd * $sum_oneside * $rate_c * $rate_a * ($dtc_a/$dtc_c)**1 * (-cos($angle_ac+2.*$angle_cc)) * $cos_a[$i-1];
}

# Calculate correlation factor by eq. 9 in main paper.
my $f_a = 1; my $f_c = 1; # f_a, f_{\tilde c} in paper
open OUT, ">convergence.dat";
print OUT "# f_a   f_c  \n";
for my $i (1..100) {
    print OUT "$i $f_a $f_c\n";
    $f_a += 2*$cos_a[$i];
    $f_c += 2*$cos_c[$i];
}
close OUT;

print "Correlation factors \n";
printf "    f_a = %.3f\n"  , $f_a;
printf "    f_c = %.3f\n\n", $f_c;

# Calculate diffusion coefficients
# Da_star: D_a^* in main paper
# Dc_star: D_{\tilde c}^* in main paper
# Da     : D(a) in main paper
# Dc     : D(c) in main paper
my $Da_star = $Cv * $eta_a * $dtc_a**2 * $f_a; # eq.6 in main paper
my $Dc_star = $Cv * $eta_c * $dtc_c**2 * $f_c; # eq.6 in main paper
my $Da = $Da_star + 2 * $Dc_star * (cos($angle_ac))**2; # eq.7 in main paper
my $Dc = 2 * $Dc_star * (cos($angle_cc))**2;            # eq.8 in main paper
print "Da*: $Da_star\n";
print "Dc*: $Dc_star\n";
print "Da : $Da\n";
print "Dc : $Dc\n\n";

print "Display diffusion coefficient [m^2/s] with periodicity M=4,6,8,10,inf\n";
for my $M (4, 6, 8, 10) {
    my $D = ($Da + $Dc) * ($M-2) / $M / 3; # D_{e-Cu3Sn} given by eq.3 in main paper
    print "  D($M): $D\n";
}
my $D = ($Da + $Dc) / 3;
print " D(inf): $D\n";
################################################
exit 0; # end of main routine
##################################################

##diffusion factor of Su
#my $Dtop1  = $Cr_top / 6 * &SC($dtc[1] , $Cv_base, $etaf2[1],  $eta_f[1]+2*$eta_f[3]);
#my $Dtop3  = $Cr_top / 6 * &SC($dtc[3] , $Cv_base, $etaf2[3],  $eta_f[3]+2*$eta_f[1]);
#my $Dbase1 = $Cr_base/ 6 * &SC($dtc[1] , $Cv_top , $etar2[1],  $eta_f[2]+$eta_r[3]+$eta_f[5]+$eta_f[10]);
#my $Dbase2 = $Cr_base/ 6 * &SC($dtc[2] , $Cv_base, $etaf2[2],  $eta_r[3]+$eta_f[5]+$eta_f[10]+$eta_r[1]);
#my $Dbase3 = $Cr_base/ 6 * &SC($dtc[3] , $Cv_top , $etaf2[3],  $eta_f[5]+$eta_f[10]+$eta_r[1]+$eta_f[2]);
#my $Dbase5 = $Cr_base/ 6 * &SC($dtc[5] , $Cv_base, $etaf2[5],  $eta_f[10]+$eta_r[1]+$eta_f[2]+$eta_r[3]);
#my $Dbase10= $Cr_base/ 6 * &SC($dtc[10], $Cv_base, $etaf2[10], $eta_r[1]+$eta_f[2]+$eta_r[3]+$eta_f[5]);
#print "Dtop1  : $Dtop1\n";
#print "Dtop3  : $Dtop3\n";
#print "Dbase1 : $Dbase1\n";
#print "Dbase2 : $Dbase2\n";
#print "Dbase3 : $Dbase3\n";
#print "Dbase5 : $Dbase5\n";
#print "Dbase10: $Dbase10\n";

sub BF { #Get boltzmann factor
    my $energy = $_[0];
    my $k_b    = 8.617342e-05;
    return exp(-$energy/($k_b*$T));
}

sub back_even{
    my $p = $_[0];
    my $max = $_[1];
    my $sum = 0;
    for (my $n=0; $n<=$max; $n+=2) {
	$sum += $p**$n * factorial(int($n+0.1)) / factorial(int(($n+0.1)/2))**2;
    }
    #printf "back_even = %.4e\n", $sum;
    return $sum;
}

sub back_odd{
    my $p = $_[0];
    my $max = $_[1];
    my $sum = 0;
    for (my $n=1; $n<=$max; $n+=2) {
	$sum += $p**$n * factorial(int($n+0.1)) / factorial(int(($n)/2)) / factorial(int(($n+1.1)/2));
    }
    #printf "back_odd = %.4e\n", $sum;
    return $sum;
}

sub back_oneside{
    my $p = $_[0];
    my $max = $_[1];
    my $sum = 0;
    for (my $n=0; $n<=$max; $n+=1){
#	print factorial(int($n+0.1))**2;
#	print "\n";
#	print  $p**(2.*$n);
#	print "\n";
	$sum += factorial(int(2*$n+0.1)) / factorial(int($n+0.1))**2 / ($n+1.) * $p**(2.*$n);
    }
    #printf "back_oneside = %.4e\n", $sum;
    return $sum;
}

#sub SC { # simple correlation
#    my $d = $_[0];
#    my $Cv= $_[1];
#    my $eta = $_[2];
#    my $eta_sum = $eta + $_[3];
##    print $eta, $eta_sum;
#    return $Cv * $d**2 * $eta * (1-2*$eta/$eta_sum);
##    return $Cv * $d**2 * $eta;
#}
